"""
Algorithm Utilities: This module contains various functions for matching graphs, which are used to save, load and
visualize the matching between longitudinal tumors.
"""

import json
import os
import warnings
from typing import Optional, List, Tuple, Union
from collections import defaultdict

import matplotlib.patheffects as PathEffects
import matplotlib.pyplot as plt
import networkx as nx


__all__ = ['get_itk_colors', 'draw_matching_graph', 'save_matching_graph', 'load_matching_graph']


def get_itk_colors(src: str = f'{os.path.dirname(__file__)}/label_descriptions.txt') -> defaultdict:
    """
    Get the colors from the ITK label descriptions file.

    Parameters
    ----------
    src : str, optional
        The path to the ITK label descriptions file. By default, it is set to the 'label_descriptions.txt' file
        in the same directory as this script.

    Returns
    -------
    defaultdict
        A dictionary where the keys are the label numbers (int) and the values are tuples representing the RGB color
        values. Each color value is a float ranging from 0.0 to 1.0. If a label number is not found in the file,
        it defaults to (0.0, 0.0, 0.0) which is black.
    """

    with open(src) as f:
        colors = f.readlines()
    colors = colors[15:]
    colors = [[c for c in l.split(' ') if c != ''] for l in colors]
    colors_default_dict = defaultdict(lambda: (0.0, 0.0, 0.0))
    for l in colors:
        colors_default_dict[int(l[0])] = (int(l[1]) / 255, int(l[2]) / 255, int(l[3]) / 255)
    return colors_default_dict


def draw_matching_graph(n_bl_nodes: int, n_fu_nodes: int, edges: List[Tuple[int, int]], case_name: str,
                        bl_weights: Optional[list] = None, fu_weights: Optional[list] = None,
                        edges_weights: Optional[dict] = None,
                        saving_file_name: Optional[str] = None, show: bool = False):
    """
    Draw a matching graph for the given tumors.

    This function prepares the data for the `_draw_matching_graph` function by creating the tumor labels and then
    calls the `_draw_matching_graph` function to draw the graph.

    Parameters
    ----------
    n_bl_nodes : int
        Number of baseline nodes.
    n_fu_nodes : int
        Number of follow-up nodes.
    edges : List[Tuple[int, int]]
        List of edges between nodes. Each edge is a tuple of two integers representing the baseline and follow-up nodes.
    case_name : str
        Name of the case. This will be displayed on the graph.
    bl_weights : Optional[list], optional
        Weights of the baseline tumors, by default None. If provided, it should contain n_bl_nodes weights. Each weight
        corresponds to a baseline tumor. The weights will be displayed on the graph.
    fu_weights : Optional[list], optional
        Weights of the follow-up tumors, by default None. If provided, it should contain n_fu_nodes weights. Each weight
        corresponds to a follow-up tumor. The weights will be displayed on the graph.
    edges_weights : Optional[dict], optional
        Weights of the edges, by default None. If provided, it should contain a dictionary where the keys are the edges
        and the values are the weights. The weights will be displayed on the graph.
    saving_file_name : Optional[str], optional
        Name of the file to save the graph, by default None.
    show : bool, optional
        Whether to show the graph, by default False.
    """

    bl_tumors = [f'{t}_bl' for t in range(1, n_bl_nodes + 1)]
    fu_tumors = [f'{t}_fu' for t in range(1, n_fu_nodes + 1)]

    _draw_matching_graph(bl_tumors, fu_tumors, n_bl_nodes, n_fu_nodes, edges, case_name, bl_weights, fu_weights,
                         edges_weights, saving_file_name, show)


def _draw_matching_graph(bl_tumors: List[str], fu_tumors: List[str], n_bl_nodes: int, n_fu_nodes: int,
                         edges: List[Tuple[int, int]], case_name: str, bl_weights: Optional[list] = None,
                         fu_weights: Optional[list] = None, edges_weights: Optional[dict] = None,
                         saving_file_name: Optional[str] = None, show: bool = False,
                         close_fig_at_end: bool = True):
    """
    Draw a matching graph for the given tumors.

    Parameters
    ----------
    bl_tumors : List[str]
        List of baseline tumors.
    fu_tumors : List[str]
        List of follow-up tumors.
    n_bl_nodes : int
        Number of baseline nodes.
    n_fu_nodes : int
        Number of follow-up nodes.
    edges : List[Tuple[int, int]]
        List of edges between nodes. Each edge is a tuple of two integers representing the baseline and follow-up nodes.
    case_name : str
        Name of the case. This will be displayed on the graph.
    bl_weights : Optional[list], optional
        Weights of the baseline tumors, by default None. If provided, it should contain n_bl_nodes weights. Each weight
        corresponds to a baseline tumor. The weights will be displayed on the graph.
    fu_weights : Optional[list], optional
        Weights of the follow-up tumors, by default None. If provided, it should contain n_fu_nodes weights. Each weight
        corresponds to a follow-up tumor. The weights will be displayed on the graph.
    edges_weights : Optional[dict], optional
        Weights of the edges, by default None. If provided, it should contain a dictionary where the keys are the edges
        and the values are the weights. The weights will be displayed on the graph.
    saving_file_name : Optional[str], optional
        Name of the file to save the graph, by default None.
    show : bool, optional
        Whether to show the graph, by default False.
    close_fig_at_end : bool, optional
        Whether to close the figure at the end, by default True.
    """

    if bl_weights is not None:
        assert len(bl_weights) == n_bl_nodes, f'the bl_weights list have to contain {n_bl_nodes} weights and ' \
                                              f'it contains {len(bl_weights)} weights: case_name = "{case_name}"'
        temp_bl_weights = dict()
        for i, bl_tumor in enumerate(bl_tumors):
            temp_bl_weights[bl_tumor] = bl_weights[i]
        bl_weights = temp_bl_weights

    if fu_weights is not None:
        assert len(fu_weights) == n_fu_nodes, f'the fu_weights list have to contain {n_fu_nodes} weights and it ' \
                                              f'contains {len(fu_weights)} weights: case_name = "{case_name}"'
        temp_fu_weights = dict()
        for i, fu_tumor in enumerate(fu_tumors):
            temp_fu_weights[fu_tumor] = fu_weights[i]
        fu_weights = temp_fu_weights

    edges = [(f'{int(e[0])}_bl', f'{int(e[1])}_fu') for e in edges]

    matched_tumors = [n for n in bl_tumors if n in (e[0] for e in edges)]
    matched_tumors += [n for n in fu_tumors if n in (e[1] for e in edges)]

    # build the graph
    G = nx.Graph()
    G.add_nodes_from(bl_tumors, bipartite='bl')
    G.add_nodes_from(fu_tumors, bipartite='fu')
    G.add_edges_from(edges)

    # define the position of the nodes
    max_range_to_draw = max(n_bl_nodes, n_fu_nodes)
    delta = 0.47 if bl_weights is not None else 0
    pos = dict(zip(bl_tumors, zip([delta] * n_bl_nodes,
                                  (max_range_to_draw * i / (n_bl_nodes - 1)
                                   for i in range(n_bl_nodes)[::-1]) if n_bl_nodes > 1 else (max_range_to_draw,))))
    delta = 0.47 if fu_weights is not None else 0
    pos.update(
        dict(zip(fu_tumors, zip([1 - delta] * n_fu_nodes,
                                (max_range_to_draw * i / (n_fu_nodes - 1)
                                 for i in range(n_fu_nodes)[::-1]) if n_fu_nodes > 1 else (max_range_to_draw,)))))

    # define the nodes' labels and colors
    nodelist = bl_tumors + fu_tumors
    nodes_labels = dict((n, int(n.split('_')[0])) for n in nodelist)
    colors = get_itk_colors()
    node_color = [colors[nodes_labels[n]] for n in nodelist]

    # draw the graph
    x = 12.4
    y = 18.8
    fig = plt.figure(figsize=(x, y))
    plt.title(f'{case_name}\n\nMatching Graph', fontsize=25)

    cf = plt.gcf()
    cf.set_facecolor("w")
    if cf._axstack is None:
        ax = cf.add_axes((0, 0, 1, 1))
    else:
        ax = cf.gca()

    ax.axis('off')

    if n_bl_nodes > 0 or n_fu_nodes > 0:

        nx.draw_networkx_nodes(G, pos, ax=ax, nodelist=nodelist, node_color=node_color, node_size=1000)
        m = y/max_range_to_draw
        c = 1
        x = c / m

        if bl_weights is not None and fu_weights is None:

            plt.text(pos[bl_tumors[0]][0] - 0.0555, pos[bl_tumors[0]][1] + x, '$_{(slice)}$ BL', fontsize=20)
            plt.text(pos[fu_tumors[0]][0] - 0.012, pos[fu_tumors[0]][1] + x, 'FU', fontsize=20)

            bl_weights_pos = dict((v, (pos[v][0] - 0.032, pos[v][1])) for v in pos if v.endswith('bl'))
            nx.draw_networkx_labels(G, bl_weights_pos, labels=bl_weights, ax=ax, font_size=15)

        elif bl_weights is None and fu_weights is not None:

            plt.text(pos[bl_tumors[0]][0] - 0.012, pos[bl_tumors[0]][1] + x, 'BL', fontsize=20)
            plt.text(pos[fu_tumors[0]][0] - 0.012, pos[fu_tumors[0]][1] + x, 'FU $_{(slice)}$', fontsize=20)

            fu_weights_pos = dict((v, (pos[v][0] + 0.032, pos[v][1])) for v in pos if v.endswith('fu'))
            nx.draw_networkx_labels(G, fu_weights_pos, labels=fu_weights, ax=ax, font_size=15)

        elif bl_weights is not None and fu_weights is not None:

            plt.text(pos[bl_tumors[0]][0] - 0.009, pos[bl_tumors[0]][1] + x, '$_{(slice)}$ BL', fontsize=20)
            plt.text(pos[fu_tumors[0]][0] - 0.0022, pos[fu_tumors[0]][1] + x, 'FU $_{(slice)}$', fontsize=20)

            bl_weights_pos = dict((v, (pos[v][0] - 0.005, pos[v][1])) for v in pos if v.endswith('bl'))
            nx.draw_networkx_labels(G, bl_weights_pos, labels=bl_weights, ax=ax, font_size=15)

            fu_weights_pos = dict((v, (pos[v][0] + 0.005, pos[v][1])) for v in pos if v.endswith('fu'))
            nx.draw_networkx_labels(G, fu_weights_pos, labels=fu_weights, ax=ax, font_size=15)

        else:

            plt.text(pos[fu_tumors[0]][0] - 0.024, pos[fu_tumors[0]][1] + x, 'FU', fontsize=20)
            plt.text(pos[bl_tumors[0]][0] - 0.024, pos[bl_tumors[0]][1] + x, 'BL', fontsize=20)

        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            nx.draw_networkx_edges(G, pos, ax=ax, edge_color='r')

            if edges_weights is not None:
                nx.draw_networkx_edge_labels(G, pos, edge_labels=edges_weights, ax=ax, label_pos=0.15, rotate=False,
                                             font_size=9)

        texts = nx.draw_networkx_labels(G, pos, ax=ax, labels=nodes_labels,
                                        # font_family='fantasy',
                                        font_size=22,
                                        font_color='w')
        for text in texts.values():
            text.set_path_effects([PathEffects.Stroke(linewidth=1, foreground='black')])

    if saving_file_name is not None:
        plt.savefig(saving_file_name)

    if show:
        # plt.show(bbox_inches='tight')
        # mng = plt.get_current_fig_manager()
        # mng.resize(*mng.window.maxsize())
        plt.show()

    if close_fig_at_end:
        plt.close(fig)


def save_matching_graph(n_bl_nodes: int, n_fu_nodes: int, edges: List[Tuple[int, int]],
                        case_name: str, saving_file_name: str, bl_weights: Optional[List[int]] = None,
                        fu_weights: Optional[List[int]] = None, bl_diameters: Optional[List[float]] = None,
                        fu_diameters: Optional[List[float]] = None, bl_organ_volume: Optional[float] = None,
                        fu_organ_volume: Optional[float] = None):
    """
    Save the matching graph data to a JSON file.

    Parameters
    ----------
    n_bl_nodes : int
        Number of baseline nodes.
    n_fu_nodes : int
        Number of follow-up nodes.
    edges : List[Tuple[int, int]]
        List of edges between nodes.
    case_name : str
        Name of the case.
    saving_file_name : str
        Name of the file to save the graph.
    bl_weights : Optional[List[int]], optional
        Weights of the baseline tumors, by default None.
    fu_weights : Optional[List[int]], optional
        Weights of the follow-up tumors, by default None.
    bl_diameters : Optional[List[float]], optional
        Diameters of the baseline tumors, by default None.
    fu_diameters : Optional[List[float]], optional
        Diameters of the follow-up tumors, by default None.
    bl_organ_volume : Optional[float], optional
        Volume of the organ at baseline, by default None.
    fu_organ_volume : Optional[float], optional
        Volume of the organ at follow-up, by default None.

    Raises
    ------
    AssertionError
        If the saving_file_name does not end with '.json'.
    """

    assert saving_file_name.endswith('.json'), 'The saving_file_name should end with ".json"'

    if bl_weights is None:
        bl_weights = []

    if fu_weights is None:
        fu_weights = []

    if bl_diameters is None:
        bl_diameters = []

    if fu_diameters is None:
        fu_diameters = []

    if bl_organ_volume is None:
        bl_organ_volume = -1

    if fu_organ_volume is None:
        fu_organ_volume = -1

    graph_dict = {'case_name': case_name, 'n_bl_nodes': n_bl_nodes, 'n_fu_nodes': n_fu_nodes, 'edges': edges,
                  'bl_weights': bl_weights, 'fu_weights': fu_weights, 'bl_diameters': bl_diameters,
                  'fu_diameters': fu_diameters, 'bl_organ_volume': bl_organ_volume, 'fu_organ_volume': fu_organ_volume}

    json_string = json.dumps(graph_dict, indent=4)
    with open(saving_file_name, "w") as json_file:
        json_file.write(json_string)


def load_matching_graph(file_name: str) -> Tuple[int, int, List[Tuple[int, int]], str, Union[List[int], None],
                                                 Union[List[int], None], Union[List[float], None],
                                                 Union[List[float], None], Union[float, None], Union[float, None]]:
    """
    Load the matching graph data from a JSON file.

    Parameters
    ----------
    file_name : str
        Path to the file to load the graph data from.

    Returns
    -------
    n_bl_nodes : int
        Number of baseline nodes.
    n_fu_nodes : int
        Number of follow-up nodes.
    edges : List[Tuple[int, int]]
        List of edges between nodes.
    case_name : str
        Name of the case.
    bl_weights : Union[List[int], None]
        Weights of the baseline tumors. If None, the baseline tumors do not have weights.
    fu_weights : Union[List[int], None]
        Weights of the follow-up tumors. If None, the follow-up tumors do not have weights.
    bl_diameters : Union[List[float], None]
        Diameters of the baseline tumors. If None, the baseline tumors diameters are not provided.
    fu_diameters : Union[List[float], None]
        Diameters of the follow-up tumors. If None, the follow-up tumors diameters are not provided.
    bl_organ_volume : Union[float, None]
        Volume of the organ at baseline. If None, the baseline organ volume is not provided.
    fu_organ_volume : Union[float, None]
        Volume of the organ at follow-up. If None, the follow-up organ volume is not provided.
    """
    with open(file_name) as json_file:
        json_string = json_file.read()
    graph_dict = json.loads(json_string)
    n_bl_nodes = graph_dict['n_bl_nodes']
    n_fu_nodes = graph_dict['n_fu_nodes']
    edges = [tuple(e) for e in graph_dict['edges']]
    case_name = graph_dict['case_name']
    bl_weights = graph_dict.get('bl_weights', None)
    fu_weights = graph_dict.get('fu_weights', None)
    bl_diameters = graph_dict.get('bl_diameters', None)
    fu_diameters = graph_dict.get('fu_diameters', None)
    bl_organ_volume = graph_dict.get('bl_organ_volume', None)
    fu_organ_volume = graph_dict.get('fu_organ_volume', None)

    if (bl_weights is not None) and (len(bl_weights) != n_bl_nodes):
        bl_weights = None

    if (fu_weights is not None) and (len(fu_weights) != n_fu_nodes):
        fu_weights = None

    if (bl_diameters is not None) and (len(bl_diameters) != n_bl_nodes):
        bl_diameters = None

    if (fu_diameters is not None) and (len(fu_diameters) != n_fu_nodes):
        fu_diameters = None

    if (bl_organ_volume is not None) and (bl_organ_volume <= 0):
        bl_organ_volume = None

    if (fu_organ_volume is not None) and (fu_organ_volume <= 0):
        fu_organ_volume = None

    return (n_bl_nodes, n_fu_nodes, edges, case_name, bl_weights, fu_weights,
            bl_diameters, fu_diameters, bl_organ_volume, fu_organ_volume)



"""
This package contains various functions for algorithm development and data analysis.
"""

from .algo_typing import *
from .data_analysis import *
from .file_operations import *
from .image_processing import *
from .matching_graphs import *
from .measurements import *
from .segmentation_features import *
from .segmentation_processing import *

__all__ = [s for s in dir() if not s.startswith('_')]



"""
Algorithm Utilities: This module contains various functions for segmentation features extraction.
"""

from typing import Tuple, Optional, Dict, List, Union

from skimage.measure import centroid
import networkx as nx
import nibabel
import numpy as np
from nibabel import affines
from scipy.ndimage import distance_transform_edt
from skimage.measure import regionprops, label

from .algo_typing import CC, IndexExpression3D, VoxelSpacing


__all__ = ['crop_to_relevant_joint_bbox', 'get_slice_of_cropped_relevant_bbox', 'fast_distance_transform_edt',
           'distance_transform_edt_for_certain_label', 'bbox2_3D', 'bbox2_2D', 'get_liver_segments',
           'get_center_of_mass', 'approximate_sphere', 'get_minimum_distance_between_CCs',
           'get_tumors_intersections', 'get_CCs_of_tumors_intersections', 'get_CCs_of_longitudinal_tumors_intersection',
           'create_approximated_spheres']


def crop_to_relevant_joint_bbox(result: np.ndarray, reference: np.ndarray) -> Tuple[np.ndarray, np.ndarray]:
    """
    This function crops the input mask arrays to the bounding box of their joint relevant region.

    Parameters
    ----------
    result : np.ndarray
        The first input mask array.
    reference : np.ndarray
        The second input mask array. It should have the same shape as the first input array.

    Returns
    -------
    Tuple[np.ndarray, np.ndarray]
        The cropped mask arrays.

    Raises
    ------
    AssertionError
        If the input arrays have different shapes.
    """
    assert result.shape == reference.shape, f'The input arrays should have the same shape. Got {result.shape} and ' \
                                            f'{reference.shape}.'
    relevant_joint_case = np.logical_or(result, reference)
    slc = get_slice_of_cropped_relevant_bbox(relevant_joint_case)
    return result[slc], reference[slc]


def get_slice_of_cropped_relevant_bbox(case: np.ndarray, margin=0) -> IndexExpression3D:
    """
    This function returns the slice of the input mask array that contains the relevant region.

    Parameters
    ----------
    case : np.ndarray
        The input mask array.
    margin : int, optional
        The margin to add to the bounding box. By default it's 0.

    Returns
    -------
    IndexExpression3D
        The slice of the cropped relevant bounding box.

    Raises
    ------
    AssertionError
        If the input margin is negative or not an integer.
    """

    assert isinstance(margin, int) and margin >= 0, f'The margin should be a non-negative integer. Got {margin}.'

    if case.ndim == 3:
        xmin, xmax, ymin, ymax, zmin, zmax = bbox2_3D(case)
        xmax += 1
        ymax += 1
        zmax += 1
        slc = np.s_[xmin: xmax, ymin: ymax, zmin: zmax]
    else:
        xmin, xmax, ymin, ymax = bbox2_2D(case)
        xmax += 1
        ymax += 1
        slc = np.s_[xmin: xmax, ymin: ymax]

    if margin > 0:
        new_slc = []
        for i in range(case.ndim):
            min_val, max_val = slc[i].start, slc[i].stop
            min_val = max(min_val - margin, 0)
            max_val = min(max_val + margin, case.shape[i])
            new_slc.append(slice(min_val, max_val))
        slc = tuple(new_slc)

    return slc


def fast_distance_transform_edt(input: np.ndarray, voxelspacing: VoxelSpacing, max_relevant_dist: float,
                                return_indices: bool) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    This function is a wrapper for the `scipy.ndimage.distance_transform_edt` function. It calculates the distance
    transform of the input mask array, faster than the original function by cropping the input array to the relevant
    region during the calculation.

    Parameters
    ----------
    input : np.ndarray
        The input mask array.
    voxelspacing : VoxelSpacing
        The voxel spacing of the input mask array.
    max_relevant_dist : float
        The maximum relevant distance to consider. If the distance is larger than this value, it will be set to
        infinity.
    return_indices : bool
        Whether to return the indices of the nearest labels or not.

    Returns
    -------
    Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]
        The distance transform of the input mask array. If `return_indices` is True, it returns a tuple containing the
        distance transform and the indices of the nearest labels.

    Raises
    ------
    AssertionError
        If the input mask array is not 2D or 3D.
    """

    input_shape = input.shape

    size_to_extend = np.ceil(max_relevant_dist / np.asarray(voxelspacing)).astype(np.int16)

    assert input.ndim in [2, 3]

    if input.ndim == 3:
        xmin, xmax, ymin, ymax, zmin, zmax = bbox2_3D(input)
    else:
        xmin, xmax, ymin, ymax = bbox2_2D(input)

    xmax += 1
    ymax += 1

    xmin = max(0, xmin - size_to_extend[0])
    xmax = min(input.shape[0], xmax + size_to_extend[0])

    ymin = max(0, ymin - size_to_extend[1])
    ymax = min(input.shape[1], ymax + size_to_extend[1])

    if input.ndim == 3:
        zmax += 1
        zmin = max(0, zmin - size_to_extend[2])
        zmax = min(input.shape[2], zmax + size_to_extend[2])

        slc = np.s_[xmin: xmax, ymin: ymax, zmin: zmax]
    else:
        slc = np.s_[xmin: xmax, ymin: ymax]

    # cropping the input image to the relevant are
    input = input[slc]

    if return_indices:
        distances, nearest_label_coords = distance_transform_edt(input == 0, return_indices=return_indices,
                                                                 sampling=voxelspacing)

        # extending the results to the input image shape
        nearest_label_coords[0] += xmin
        nearest_label_coords[1] += ymin
        if input.ndim == 3:
            nearest_label_coords[2] += zmin

        extended_nearest_label_coords = np.zeros((input.ndim,) + input_shape)
        if input.ndim == 3:
            extended_nearest_label_coords[:, slc[0], slc[1], slc[2]] = nearest_label_coords
        else:
            extended_nearest_label_coords[:, slc[0], slc[1]] = nearest_label_coords

    else:
        distances = distance_transform_edt(input == 0, return_indices=return_indices, sampling=voxelspacing)

    # extending the results to the input image shape
    extended_distances = np.ones(input_shape) * np.inf
    extended_distances[slc] = distances

    if return_indices:
        return extended_distances, extended_nearest_label_coords

    return extended_distances


def distance_transform_edt_for_certain_label(label_and_max_relevant_dist: Tuple[int, float], label_image: np.ndarray,
                                             voxelspacing: VoxelSpacing,
                                             return_indices: bool) -> Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]:
    """
    This function calculates the distance transform of the input mask array for a certain label.

    Parameters
    ----------
    label_and_max_relevant_dist: Tuple[int, float]
        A tuple containing the label to calculate the distance transform for and the maximum relevant distance to
        consider. If the distance is larger than this value, it will be set to infinity.
    label_image: np.ndarray
        The input labeled mask array.
    voxelspacing: VoxelSpacing
        The voxel spacing of the input mask array.
    return_indices: bool
        Whether to return the indices of the nearest labels or not.

    Returns
    -------
    Union[np.ndarray, Tuple[np.ndarray, np.ndarray]]
        The distance transform of the input mask array for the given label. If `return_indices` is True, it returns a
        tuple containing the distance transform and the indices of the nearest labels.
    """
    label, max_relevant_dist = label_and_max_relevant_dist
    return fast_distance_transform_edt(label_image == label, voxelspacing, max_relevant_dist,
                                       return_indices=return_indices)


def bbox2_3D(img: np.ndarray) -> Tuple[int, ...]:
    """
    This function calculates the bounding box of the input 3D mask array.

    Notes
    -----
    The maximum values are inclusive.

    Parameters
    ----------
    img : np.ndarray
        The input 3D mask array.

    Returns
    -------
    xmin, xmax, ymin, ymax, zmin, zmax : Tuple[int, ...]
        The bounding box of the input 3D mask array.
    """
    x = np.any(img, axis=(1, 2))
    y = np.any(img, axis=(0, 2))
    z = np.any(img, axis=(0, 1))

    xmin, xmax = np.where(x)[0][[0, -1]]
    ymin, ymax = np.where(y)[0][[0, -1]]
    zmin, zmax = np.where(z)[0][[0, -1]]

    return xmin, xmax, ymin, ymax, zmin, zmax


def bbox2_2D(img):
    """
    This function calculates the bounding box of the input 2D mask array.

    Notes
    -----
    The maximum values are inclusive.

    Parameters
    ----------
    img : np.ndarray
        The input 2D mask array.

    Returns
    -------
    xmin, xmax, ymin, ymax : Tuple[int, ...]
        The bounding box of the input 2D mask array.
    """
    x = np.any(img, axis=1)
    y = np.any(img, axis=0)

    xmin, xmax = np.where(x)[0][[0, -1]]
    ymin, ymax = np.where(y)[0][[0, -1]]

    return xmin, xmax, ymin, ymax


def get_liver_segments(liver_case: np.ndarray) -> np.ndarray:
    """
    This function segments the liver into 3 segments based on the bounding box of the input 3D mask array. The segments
    are as follows: 1 for the upper left, 2 for the upper right, and 3 for the lower left.

    Parameters
    ----------
    liver_case : np.ndarray
        The input 3D liver mask array.

    Returns
    -------
    np.ndarray
        The segmented liver mask array.
    """
    xmin, xmax, ymin, ymax, _, _ = bbox2_3D(liver_case)
    res = np.zeros_like(liver_case)
    res[(xmin + xmax) // 2:xmax, (ymin + ymax) // 2:ymax, :] = 1
    res[xmin:(xmin + xmax) // 2, (ymin + ymax) // 2:ymax, :] = 2
    res[:, ymin:(ymin + ymax) // 2, :] = 3
    res *= liver_case
    return res


def get_center_of_mass(img: np.ndarray) -> Tuple[int, int, int]:
    """
    Returns the center of mass of the input 3D mask array. The center of mass is calculated as the center of the mask's
    bounding box.

    Parameters
    ----------
    img : np.ndarray
        The input 3D mask array.

    Returns
    -------
    Tuple[int, int, int]
        The center of mass of the input 3D mask array.
    """
    xmin, xmax, ymin, ymax, zmin, zmax = bbox2_3D(img)
    return (xmin + xmax) // 2, (ymin + ymax) // 2, (zmin + zmax) // 2


def approximate_sphere(relevant_points_in_real_space: np.ndarray, relevant_points_in_voxel_space: np.ndarray,
                       center_of_mass_in_voxel_space: Tuple[int, int, int], approximate_radius_in_real_space: float,
                       affine_matrix: np.ndarray):
    """
    This function approximates a sphere around the center of mass of the input points.

    Parameters
    ----------
    relevant_points_in_real_space : np.ndarray
        The relevant points in real space. It should be a 2D array with shape (n, 3).
    relevant_points_in_voxel_space : np.ndarray
        The relevant points in voxel space. It should be a 2D array with shape (n, 3).
    center_of_mass_in_voxel_space : Tuple[int, int, int]
        The center of mass of the input points in voxel space.
    approximate_radius_in_real_space : float
        The approximate radius of the sphere in real space.
    affine_matrix : np.ndarray
        The affine matrix that maps the voxel space to the real space.

    Returns
    -------
    np.ndarray
        The approximated sphere mask array.
    """
    center_of_mass_in_real_space = affines.apply_affine(affine_matrix, center_of_mass_in_voxel_space)
    final_points_in_voxel_space = relevant_points_in_voxel_space[
        ((relevant_points_in_real_space - center_of_mass_in_real_space) ** 2).sum(
            axis=1) <= approximate_radius_in_real_space ** 2]
    sphere = np.zeros(relevant_points_in_voxel_space[-1] + 1)
    sphere[final_points_in_voxel_space[:, 0], final_points_in_voxel_space[:, 1], final_points_in_voxel_space[:, 2]] = 1
    return sphere


def get_minimum_distance_between_CCs(mask: np.ndarray, voxel_to_real_space_trans: Optional[np.ndarray] = None,
                                     max_points_per_CC: Optional[int] = None, seed: Optional[int] = None,
                                     connectivity: Optional[int] = None) -> float:
    """
    Get the minimum distance between every 2 connected components in a binary image. The distance is calculated as the
    minimum distance between every 2 connected components' points. If the number of points in a connected component is
    larger than `max_points_per_CC`, a random subset of `max_points_per_CC` points will be considered. If the affine
    matrix `voxel_to_real_space_trans` is given, the points will be transformed to the real space. The distance is
    calculated as the Euclidean distance between the points.

    Notes
    -----
    If the input binary image contains only one connected component, the function will return infinity.

    Parameters
    ----------
    mask : np.ndarray
        The input binary image.
    voxel_to_real_space_trans : np.ndarray, optional
        The affine matrix that maps the voxel space to the real space. By default it's None.
    max_points_per_CC : int, optional
    seed : int, optional
        The seed to use for the random number generator. By default it's None.
    connectivity : int, optional
        The connectivity to consider. By default it's None.

    Returns
    -------
    float
        The minimum distance between every 2 connected components in the input binary image.
    """

    rand = np.random.RandomState(seed)

    def dist_between_PCs_squared(pc1, pc2):
        return np.min(((np.expand_dims(pc1, -1) - np.expand_dims(pc2.T, 0)) ** 2).sum(axis=1))

    def choose_n_random_points(pc: np.ndarray, n) -> np.ndarray:
        perm = rand.permutation(pc.shape[0])
        idx = perm[:n]
        return pc[idx, :]

    def filter_and_transfer_points(pc: np.ndarray):
        if (max_points_per_CC is not None) and (pc.shape[0] > max_points_per_CC):
            pc = choose_n_random_points(pc, max_points_per_CC)
        if voxel_to_real_space_trans is not None:
            pc = affines.apply_affine(voxel_to_real_space_trans, pc)
        return pc

    mask = label(mask > 0, connectivity=connectivity)

    list_of_PCs = [filter_and_transfer_points(r.coords) for r in regionprops(mask)]

    n_CCs = len(list_of_PCs)

    if n_CCs >= 2:
        return np.sqrt(np.min([dist_between_PCs_squared(list_of_PCs[i], list_of_PCs[j])
                               for i in range(n_CCs) for j in range(i + 1, n_CCs)]))

    return np.inf


def get_tumors_intersections(gt: np.ndarray, pred: np.ndarray,
                             unique_intersections_only: bool = False) -> Dict[int, List[int]]:
    """
    Get intersections of tumors between GT and PRED labeled masks.

    Parameters
    ----------
    gt : np.ndarray
        Labeled GT tumors case.
    pred : np.ndarray
        Labeled PRED tumors case.
    unique_intersections_only : bool, False by default
        If considering only unique intersections.

    Returns
    -------
    Dict[int, List[int]]
        A dict containing for each relevant GT tumor (key) a list with the relevant intersections (value).
    """

    # extract intersection pairs of tumors
    pairs = np.hstack([gt.reshape([-1, 1]), pred.reshape([-1, 1])])
    pairs = np.unique(pairs[~np.any(pairs == 0, axis=1)], axis=0)

    if unique_intersections_only:
        # filter out unique connections
        unique_gt = np.stack(np.unique(pairs[:, 0], return_counts=True)).T
        unique_gt = unique_gt[unique_gt[:, 1] == 1][:, 0]
        unique_pred = np.stack(np.unique(pairs[:, 1], return_counts=True)).T
        unique_pred = unique_pred[unique_pred[:, 1] == 1][:, 0]
        pairs = pairs[np.isin(pairs[:, 0], unique_gt)]
        pairs = pairs[np.isin(pairs[:, 1], unique_pred)]

    intersections = []
    previous_gt = None
    for k, gt in enumerate(pairs[:, 0]):
        if previous_gt is not None and (gt == previous_gt[0]):
            previous_gt = (gt, previous_gt[1] + [int(pairs[k, 1])])
            intersections[-1] = previous_gt
        else:
            previous_gt = (gt, [int(pairs[k, 1])])
            intersections.append(previous_gt)

    return dict(intersections)


def get_CCs_of_tumors_intersections(gt: np.ndarray, pred: np.ndarray) -> List[CC]:
    """
    Get Connected Components (CC) intersections of tumors between GT and PRED.

    Parameters
    ----------
    gt : numpy.ndarray
        Labeled GT tumors case.
    pred : numpy.ndarray
        Labeled PRED tumors case.

    Returns
    -------
    CCs : list of CCs
        List of CCs. Each CC is a tuple of two list of ints, indicating the gt and pred tumors in the corresponding CC,
        respectively.
    """

    # extract intersection tumors
    intersection_tumors = get_tumors_intersections(gt, pred)

    # define nodes
    define_nodes = lambda ts, extension: [f'{int(t)}{extension}' for t in np.unique(ts) if t != 0]
    gt_nodes = define_nodes(gt, '_gt')
    pred_nodes = define_nodes(pred, '_pred')

    # define edges
    edges = []
    for gt_t in intersection_tumors:
        for pred_t in intersection_tumors[gt_t]:
            edges.append((f'{int(gt_t)}_gt', f'{int(pred_t)}_pred'))

    # build the graph
    G = nx.Graph()
    G.add_nodes_from(gt_nodes, bipartite='gt')
    G.add_nodes_from(pred_nodes, bipartite='pred')
    G.add_edges_from(edges)

    # extract CCs
    CCs = []
    for cc in nx.algorithms.components.connected_components(G):
        current_gts, current_preds = [], []
        for t in cc:
            _t = int(t.split('_')[0])
            if t.endswith('_gt'):
                current_gts.append(_t)
            else:
                current_preds.append(_t)
        CCs.append((current_gts, current_preds))

    return CCs


def get_CCs_of_longitudinal_tumors_intersection(tumors_masks: List[np.ndarray]):
    """
    Get Connected Components (CC) intersections of tumors between multiple tumors masks arrays.

    Parameters
    ----------
    tumors_masks : list of ndarray
        A list containing arrays containing labeled tumors masks. The list should contain at least two arrays.

    Returns
    -------
    CCs : list of CCs
        List of CCs. Each CC is a tuple of `n` list of ints, where `n` is the size of the given `tumors_mask` list,
        indicating the tumors in the corresponding CC, respectively.

    Raises
    ------
    AssertionError
        If the input tumors masks list is empty or contains only one element.
    """

    assert len(tumors_masks) >= 2

    define_nodes = lambda ts, extension: [f'{int(t)}{extension}' for t in np.unique(ts) if t != 0]

    # build the graph
    G = nx.Graph()

    # add 1st array lesion nodes
    G.add_nodes_from(define_nodes(tumors_masks[0], '_0'), bipartite='0')

    for i in range(len(tumors_masks) - 1):
        # extract intersection tumors
        intersection_tumors = get_tumors_intersections(tumors_masks[i], tumors_masks[i + 1])

        # define paired array lesion nodes
        new_nodes = define_nodes(tumors_masks[i + 1], f'_{i + 1}')

        # define edges
        edges = []
        for gt_t in intersection_tumors:
            for pred_t in intersection_tumors[gt_t]:
                edges.append((f'{int(gt_t)}_{i}', f'{int(pred_t)}_{i + 1}'))

        # add new nodes and edges to the graph
        G.add_nodes_from(new_nodes, bipartite=f'{i + 1}')
        G.add_edges_from(edges)

    # extract CCs
    CCs = []
    for cc in nx.algorithms.components.connected_components(G):
        current_ccs = []
        for _ in range(len(tumors_masks)):
            current_ccs.append([])
        for t in cc:
            label, ind = [int(c) for c in t.split('_')]
            current_ccs[ind].append(label)
        current_ccs = tuple([sorted(ccs) for ccs in current_ccs])
        CCs.append(current_ccs)

    return CCs


def create_approximated_spheres(labeled_tumors_mask: nibabel.Nifti1Image,
                                desired_tumors_labels: Optional[List[int]] = None) -> np.ndarray:
    """
    Creates approximated spheres for a given labeled tumors masks.

    Parameters
    ----------
    labeled_tumors_mask : nibabel.Nifti1Image
        The desired labeled tumors mask nifti file as a nibabel.Nifti1Image object.
    desired_tumors_labels : list of int, optional
        The desired tumor labels to consider. If None, all the tumors will be considered.

    Returns
    -------
    res : numpy.ndarray
        An ndarray same shape as given mask with all the tumors (or the desired ones) replaced with approximated
        spheres.
    """

    tumors = np.round(labeled_tumors_mask.get_fdata(dtype=np.float32))
    pix_dims = labeled_tumors_mask.header.get_zooms()
    voxel_volume = pix_dims[0] * pix_dims[1] * pix_dims[2]
    affine_matrix = labeled_tumors_mask.affine

    nX, nY, nZ = tumors.shape
    min_p = affines.apply_affine(affine_matrix, (0, 0, 0))
    max_p = affines.apply_affine(affine_matrix, tumors.shape)
    relevant_points_in_real_space = np.vstack([np.repeat(np.arange(min_p[0], max_p[0], pix_dims[0]), nY * nZ),
                                               np.tile(np.repeat(np.arange(min_p[1], max_p[1], pix_dims[1]), nZ), nX),
                                               np.tile(np.arange(min_p[2], max_p[2], pix_dims[2]), nX * nY)]).T
    relevant_points_in_voxel_space = np.vstack([np.repeat(np.arange(0, nX), nY * nZ),
                                                np.tile(np.repeat(np.arange(0, nY), nZ), nX),
                                                np.tile(np.arange(0, nZ), nX * nY)]).T

    tumors_to_consider = np.unique(tumors) if desired_tumors_labels is None else desired_tumors_labels
    if desired_tumors_labels is None:
        tumors_to_consider = list(tumors_to_consider[tumors_to_consider != 0])

    res = np.zeros_like(tumors)
    for t in tumors_to_consider:
        current_t = (tumors == t).astype(tumors.dtype)
        current_t_vol = current_t.sum() * voxel_volume
        current_t_diameter = 2 * ((3 * current_t_vol) / (4 * np.pi)) ** (1 / 3) # diameter = 2 * (3V / 4pi)^(1/3)
        current_t_centroid = centroid(current_t)
        current_sphere = approximate_sphere(relevant_points_in_real_space, relevant_points_in_voxel_space,
                                            current_t_centroid, current_t_diameter / 2, affine_matrix)
        res[current_sphere == 1] = 1

    return res
